/***********************************************************************
|  Permutation group analyzer
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* normsub.c: normal subgroups and factor groups */

#include "grpnode.h"

static short tp1[PERMSIZE], tp2[PERMSIZE];
static char firstfactor, firstabel;

/* turn coset reps into a single list of generators */
static cosrep_to_list(l)
struct list_perm *l;
{
int n;
struct perm *p;

init_0(l);
for(n=0; n<permsize; ++n)
LOOP_perm(&cosrep[n], p)
extract_perm(&cosrep[n], p);
append_perm(l, p);
END_LOOP_perm
} /* cosrep_to_list */

/* turn list of generators into strong generators, the pseudo inverse of the above */
static make_stronggen(l)
struct list_perm *l;
{
struct perm *p;
LOOP_perm(l, p)
foldin_gen(p->p, 0);
extract_free_perm(l, p);
END_LOOP_perm
} /* make_stronggen */

/* free up coset reps */
static zap_cosrep()
{
struct list_perm l;
cosrep_to_list(&l);
zap_list_perm(&l);
} /* zap_cosrep */

/* renumber the elements of the permutation group */
static renumber(l, remap)
struct list_perm *l;
short *remap;
{
int i;
struct perm *p;

LOOP_perm(l, p)
for(i=0; i<permsize; ++i)
if(remap[i] >= 0)
tp1[remap[i]] = remap[p->p[i]];
permcpy(p->p, tp1);
END_LOOP_perm
} /* renumber */

static abelprint(n, exp)
{
while(exp--) {
if(!firstabel) printf("+");
else {
if(!firstfactor) printf(".");
firstfactor = 0;
firstabel = 0;
}
printf("Z%d",n);
}
} /* ableprint */

/* decompose an abelian group */
static abelcheck(l)
struct list_perm *l;
{
struct perm *a, *b;
int i, j, n, p, nfactors, base, fudge, x, y, z;
short gorder;  /* group order */
short numord[PERMSIZE+1]; /* number of elements of each order */
/* power of a prime */
static char ispp[PERMSIZE+1] = {
0,0,2,3,2,5,0,7,2,3,
0,11,0,13,0,0,2,17,0,19,
0,0,0,23,0,5,0,3,0,29,
0,31,2,0,0,0,0,37,0,0,
0,41,0,43,0,0,0,47,0,7,
0,0,0,53,0,0,0,0,0,59,
0,61,0,0,2,0,0,67,0,0,
0,71,0,73,0,0,0,0,0,79,
0,3,0,83,0,0,0,0,0,89,
0,0,0,0,0,0,0};

/* make sure we've got an abelian group on our hands */
LOOP_perm(l, a)
RANGE_LOOP_perm(b, a, 0)
if(!theycommute(a->p,b->p))
return 0;
END_LOOP_perm
END_LOOP_perm

make_stronggen(l);
firstabel = 1;

for(n=permsize-1; n>=0; --n) {
gorder = cosrep[n].n_nodes + 1;
if(gorder == 1) continue;

/* do not think that this last group is cyclic.
 * gorder only gives us the order of the group.
 * This is illustrated by 0123 1032 2301 3210 (Z2+Z2).
 * So we decompose the group by looking at the order of the individual elements.
 * We're not using ->tostart, so let it hold the order of the element. */
memset(numord,0,sizeof(numord));
LOOP_perm(&cosrep[n], a)
i = get_cyclen(a->p);
if(ispp[i]) ++numord[i];
END_LOOP_perm

for(p=2; p<=permsize; ++p) {
if(ispp[p] != p) continue;
base = numord[p];
if(!base) continue;

/* if j separate factors based on powers of p,
 * then p^j-1 elements will have order p */
++base;
nfactors = 0;
while(base > 1)
base /= p, ++nfactors;
fudge = 1;

/* look at higher powers of p */
for(i=p*p; i<=permsize; i*=p) {
base = numord[i];
if(!base) break;

/* if j factors extend to i, the next power of p,
 * then the number of elements of order i is i^j*(i/p)^(nfactors-j) -
 * (i/p)^j * (i/p)^(nfactors-j). */
x = y = z = 1;
for(j=0; j<nfactors; ++j) z *= i/p;

for(j=1; j<=nfactors; ++j) {
x *= i/p;
z /= i/p;
y *= i;
if(base == (y-x) * z * fudge) break;
}
if(j > nfactors) eprint("@j exceeds nfactors");

abelprint(i/p, nfactors-j);
fudge *= z;
nfactors = j;
} /* loop over powers of p */

abelprint(i/p,nfactors);
} /* loop over primes */

/* normalize all other (upper) coset reps */
for(i=0; i<n; ++i) {
LOOP_perm(&cosrep[i], a)
if(a->p[n] == n) continue;
b = find_cosrep(n, a->p[n], 0, 1);
apply(a->p, b->p, tp1, 1);
permcpy(a->p, tp1);
if(tp1[n] != n)
eprint("@ablcheck can't divide by bottom group");
END_LOOP_perm
} /* loop dividing by the bottom group */
} /* loop going up the levels of the strong generators */

zap_cosrep();
return 1;
} /* abelcheck */

/* remove elements that never move */
static unmoving(l)
struct list_perm *l;
{
int i, j;
struct perm *p;
short remap[PERMSIZE];

for(i=j=0; i<permsize; ++i) {
remap[i] = -1; /* skip flag */
LOOP_perm(l, p)
if(p->p[i] == i) continue;
/* p moves i, and that's all we need to know */
remap[i] = j++;
break;
END_LOOP_perm
}

if(j < permsize) {
renumber(l, remap);
permsize = j;
}
} /* unmoving */

/* if group is not transitive, divide and conquer */
static untransitive(l)
struct list_perm *l;
{
int i, j, n, change;
struct perm *p;
short remap[PERMSIZE];

memset(remap, -1, sizeof(remap));
remap[0] = 1;
j = 1;
change = 1;

while(change) {
change = 0;
LOOP_perm(l, p)
for(i=0; i<permsize; ++i)
if(remap[i] > 0 && remap[p->p[i]] < 0) {
remap[p->p[i]] = 1;
change = 1;
++j;
}
END_LOOP_perm
}
if(j == permsize) return -1; /* already transitive */

/* now renumber elements, insiders first, then elements outside the orbit of 0 */
n = 0;
for(i=0; i<permsize; ++i)
if(remap[i] > 0) remap[i] = n++;
for(i=0; i<permsize; ++i)
if(remap[i] < 0) remap[i] = n++;

renumber(l, remap);
return j;
} /* untransitive */

decompose()
{
static struct list_perm l;
struct perm *p;
int i;

printf("Decomposition series: ");
for(i=0; i<num_moves; ++i) {
p = get_perm();
permcpy(p->p, moveperm[i]);
append_perm(&l, p);
}

firstfactor = 1;
factor(&l);
if(firstfactor) printf("E");
nl();
} /* decompose */

/* split a group into kernel and quotient */
/* frees the passed-in group */
static factor(l)
struct list_perm *l;
{
struct list_perm new_l;
struct perm *p;
int i, n;
int save_permsize;
short remap[PERMSIZE];

unmoving(l);
if(!permsize) return;

if(abelcheck(l)) return;

untrans:
save_permsize = permsize;
n = untransitive(l);
if(n >= 0) {
copy_list_perm(&new_l, l);
permsize = n;
factor(&new_l);

permsize = save_permsize;
make_stronggen(l);

/* keep only the generators that fix the first n */
for(i=0; i<n; ++i)
zap_list_perm(&cosrep[i]);
cosrep_to_list(l);

memset(remap, -1, sizeof(remap));
for(i=n; i<permsize; ++i) remap[i] = i-n;
renumber(l, remap);
permsize -= n;
factor(l);
return;
}

/* now try to split elements into blocks */
for(i=1; i<permsize; ++i) {
permcpy(remap, idn->p);
remap[i] = 0;
if(block_closure(l, remap) >= 0) goto untrans;
} /* loop trying to split into blocks */

/* not analyzable, just show the order */
make_stronggen(l);
if(!firstfactor) printf(".");
firstfactor = 0;

/* test for An and Sn */
for(i=0; i<permsize; ++i)
if(cosrep[i].n_nodes != permsize-i-1) break;
if(i == permsize) { printf("S%d", permsize); goto done; }
if(i == permsize-2) { printf("A%d", permsize); goto done; }

/* not analyzable by this program */
printf("?");
printord1();

done:
zap_cosrep();
} /* factor */

static block_closure(l, remap)
struct list_perm *l;
short *remap;
{
int i, j, i1, j1, k, swap, change;
struct perm *p;

change = 1;
while(change) {
change = 0;
for(i=0; i<permsize; ++i) {
if(remap[i] < 0) continue;
for(j=i+1; j<permsize; ++j) {
if(remap[j] != remap[i]) continue;
/* two members of the same block, always mapped to the same block */
LOOP_perm(l, p)
i1 = remap[p->p[i]];
j1 = remap[p->p[j]];
if(i1 == j1) continue; /* already equal */
if(j1 < i1) swap = i1, i1 = j1, j1 = swap;
for(k=0; k<permsize; ++k)
if(remap[k] == j1) remap[k] = i1;
change = 1;
END_LOOP_perm
}
}
}

/* if everything equates to element 0 then we have no block decomposition */
for(i=j=0; i<permsize; ++i)
if(remap[i] == 0) ++j;
if(j == permsize) return -1;

j1 = j;
i1 = permsize/j1;
if(i1+permsize > PERMSIZE) eprint("1internal group requires %d elements, limit %d", permsize+i1, PERMSIZE);
for(i=0; i<i1; ++i) {
for(j=0; j<permsize; ++j)
if(remap[j] >= i) break;
if(j == permsize) eprint("@could not find block %d", i);
j = remap[j];
for(k=0; k<permsize; ++k)
if(remap[k] == j) remap[k] = i;
}

/* extend the permutations to describe block rearrangement */
LOOP_perm(l, p)
for(i=0; i<permsize; ++i)
p->p[permsize+remap[i]] = remap[p->p[i]] + permsize;
END_LOOP_perm

for(i=0; i<i1; ++i)
remap[i+permsize] = i;
for(i=0; i<permsize; ++i)
remap[i] = i+i1;
permsize += i1;
renumber(l, remap);
return i1;
} /* block_closure */

